#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#include <oslib/displayfield.h>
#include <oslib/osfile.h>
#include <oslib/osfscontrol.h>
#include <oslib/osspriteop.h>
#include <oslib/osgbpb.h>
#include <oslib/quit.h>
#include <oslib/stringset.h>
#include <oslib/writablefield.h>
#include <oslib/window.h>
#include <oslib/wimp.h>
#include <oslib/colourtrans.h>
#include <oslib/iconbar.h>
#include <oslib/hourglass.h>
#include <oslib/hal.h>
#include <oslib/saveas.h>
#include <oslib/menu.h>
#include <oslib/actionbutton.h>
#include <oslib/radiobutton.h>
#include <Systypes.h>
#include <Event.h>
#include <MagickWand/MagickWand.h>

#include <unixlib/local.h>

static int toolbox_quit;
static messagetrans_control_block messages_file;
static toolbox_o main_window;
static wimp_w w;
static int width, height;
static int converting;
static osspriteop_area *spritearea = NULL;
static osspriteop_id header;
static int size;
static char imagefile[256];
static double startTime;
static double endTime;

toolbox_action_list toolbox_events;     // holds toolbox event codes
wimp_message_list wimp_messages;

static void toolbox_poll(int active);

typedef struct toolbox_message_list {
	int first;
	int rest[];
} toolbox_message_list;

typedef union message_list {
	toolbox_message_list toolbox;
	wimp_message_list wimp;
	toolbox_action_list Toolbox;
} message_list;

typedef enum toolbox_result {
	TOOLBOX_RESULT_OK,
	TOOLBOX_RESULT_NOMEM,
	TOOLBOX_RESULT_OPEN,
	TOOLBOX_RESULT_WRITE
} toolbox_result;

#ifndef UNUSED
#define UNUSED(x) ((x)=(x))
#endif

#define PTR_WIMP_MESSAGE_LIST(l) ((wimp_message_list *) (message_list *) (l))
#define PTR_TOOLBOX_ACTION_LIST(l) ((toolbox_action_list *) (message_list *) (l))


/* Wimp Messages we're interested in */
/* static toolbox_message_list messages = {
	message_DATA_LOAD,
	{
		message_QUIT,
		0
	}
}; */

/* Toolbox events we're interested in */
/* static toolbox_message_list tbcodes = {
	Convert_Font,
	{
		action_ERROR,
		action_QUIT_QUIT,
		0
	}
};  */

static void toolbox_init(int argc, char **argv);
static void toolbox_exit(void);
static void register_toolbox_handlers(void);
static osbool toolbox_event_quit(bits event_code, toolbox_action *event,toolbox_block *id_block, void *handle);
static osbool redraw_window(wimp_event_no event_code,wimp_block *block, toolbox_block *id_block,void *handl);
static osbool toolbox_event_error(bits event_code, toolbox_action *event,
		toolbox_block *id_block, void *handle);
static void register_message_handlers(void);
static int wimp_message_quit(wimp_message *message,void *handle);
static int wimp_message_dataload(wimp_message *message,void *handle);
static void bitmap_load(char *filename, size_t resize_width, size_t resize_height);
static int bitmap_convert(char *filename, int file_type, char *output_file, size_t resize_width, size_t resize_height);
static osbool close_window(wimp_event_no event_code, wimp_block *block, toolbox_block *id_block, void *handle);
static void destroy_window(void);
extern double get_timerValue(void);
osbool toolbox_saveas(bits event_code, toolbox_action *event, toolbox_block *id_block, void *handle);
osbool toolbox_savetofile(bits event_code, toolbox_action *event, toolbox_block *id_block, void *handle);
osbool toolbox_actionbutton(bits event_code, toolbox_action *event, toolbox_block *id_block, void *handle);
osbool update_menu(bits event_code, toolbox_action *block, toolbox_block *id_block, void *handle);

#define resizemenu_CANCEL 0x04
#define resizemenu_RESIZE 0x07

int main(int argc, char **argv)
{
	toolbox_init(argc, argv);

	while (!toolbox_quit)
		toolbox_poll(0);

	toolbox_exit();

	return 0;
}


void toolbox_init(int argc, char **argv)
{
  os_error *error;

	toolbox_block toolbox_block;
	osbool success;

    wimp_messages.messages[0]=0;
    toolbox_events.action_nos[0]=0;

	UNUSED(argc);
	UNUSED(argv);

	event_initialise(&toolbox_block);

	event_set_mask(wimp_MASK_NULL |
			wimp_MASK_LEAVING |
			wimp_MASK_ENTERING |
			wimp_QUEUE_MOUSE |
			wimp_QUEUE_KEY |
			wimp_MASK_LOSE |
			wimp_MASK_GAIN);

	register_toolbox_handlers();
	register_message_handlers();


	error = xtoolbox_initialise(0, 310,
				&wimp_messages,
				&toolbox_events,
				"<ImageView$Dir>", &messages_file,
				&toolbox_block, NULL, NULL, NULL);
	if (error) {
		fprintf(stderr, "toolbox_initialise: 0x%x: %s\n",
			error->errnum, error->errmess);
			exit(1);
	}


	error = xtoolbox_create_object(0, (toolbox_id) "Window", &main_window);
	if (error) {
		fprintf(stderr, "toolbox_create_object: 0x%x: %s\n",
			error->errnum, error->errmess);
		 exit(1);


	  }
	success = event_register_wimp_handler(event_ANY,
			wimp_REDRAW_WINDOW_REQUEST, redraw_window, NULL);
	if (success == FALSE)
	fprintf(stderr, "registering RedrawWindow failed\n");

	success = event_register_wimp_handler(event_ANY,
			wimp_CLOSE_WINDOW_REQUEST, close_window, NULL);
	if (success == FALSE)
	fprintf(stderr, "registering CloseWindow failed\n");

}

static void bitmap_load(char *filename, size_t resize_width, size_t resize_height)
{
  os_error *error;
  BBox extent;
  osbool success;

	 if (main_window)
	   destroy_window();

	   xhourglass_on();
	   startTime = get_timerValue();

	error = xtoolbox_create_object(0, (toolbox_id) "Window", &main_window);
	if (error) {
		fprintf(stderr, "toolbox_create_object: 0x%x: %s\n",
			error->errnum, error->errmess);
		 exit(1);


	  }
		error = xwindow_set_title(0, main_window, filename);
	if (error != NULL) {
		fprintf(stderr, "window_set_title: 0x%x: %s\n",
			error->errnum, error->errmess);
	}

	success = event_register_wimp_handler(event_ANY,
			wimp_REDRAW_WINDOW_REQUEST, redraw_window, NULL);
	if (success == FALSE)
	fprintf(stderr, "registering RedrawWindow failed\n");

	success = event_register_wimp_handler(event_ANY,
			wimp_CLOSE_WINDOW_REQUEST, close_window, NULL);
	if (success == FALSE)
	fprintf(stderr, "registering CloseWindow failed\n");

	if (bitmap_convert(filename, 0xff9, NULL, resize_width, resize_height) == 0)
	{
	   xwindow_get_extent(0, main_window, &extent);
	   extent.y0 = - height * 2;
	   extent.x1 = width * 2;
	   xwindow_set_extent(0, main_window, &extent);
	   fprintf(stderr, "x0: %d y0: %d x1: %d y1: %d\n", extent.x0, extent.y0, extent.x1, extent.y1);
	   xtoolbox_show_object(0, main_window, toolbox_POSITION_DEFAULT, 0, NULL, (toolbox_c)NULL);
	}

	endTime = get_timerValue();
	fprintf(stderr, "Process time total %fs for %s\n", endTime - startTime, filename);

  xhourglass_off();

}

static void toolbox_poll(int active)
{
	int event;
	unsigned int mask;
	wimp_block block;

	if (active || converting) {
		event_set_mask(0x3972);
		event_poll(&event, &block, 0);
	} else {
		event_get_mask(&mask);
		event_set_mask(mask | wimp_MASK_NULL);
		event_poll(&event, &block, 0);
	}
}

void toolbox_exit(void)
{
	if (spritearea) {
	  free(spritearea);
	  spritearea = NULL;
	  }

}

/**
 * Register event handlers with the toolbox
 */
void register_toolbox_handlers(void)
{
	 osbool success;


 	success = event_register_toolbox_handler(event_ANY,
			action_QUIT_QUIT, toolbox_event_quit, NULL);
	if (success == FALSE)
		fprintf(stderr, "registering quit_QUIT failed\n");

	success = event_register_toolbox_handler(event_ANY,
			action_ERROR, toolbox_event_error, NULL);
	if (success == FALSE)
		fprintf(stderr, "registering action_ERROR failed\n");
	success = event_register_toolbox_handler(event_ANY,
			action_SAVE_AS_ABOUT_TO_BE_SHOWN, toolbox_saveas, NULL);
	if (success == FALSE)
		fprintf(stderr, "registering action_SAVE_AS_ABOUT_TO_BE_SHOWN failed\n");
	success = event_register_toolbox_handler(event_ANY,
			action_SAVE_AS_SAVE_TO_FILE, toolbox_savetofile, NULL);
	if (success == FALSE)
		fprintf(stderr, "registering action_SAVE_AS_SAVE_TO_FILE failed\n");
	success = event_register_toolbox_handler(event_ANY,
			resizemenu_RESIZE, toolbox_actionbutton, NULL);
	if (success == FALSE)
	fprintf(stderr, "registering resizemenu_RESIZE failed\n");
		success = event_register_toolbox_handler(event_ANY,
			action_WINDOW_ABOUT_TO_BE_SHOWN, update_menu, NULL);
	if (success == FALSE)
		fprintf(stderr, "registering action_WINDOW_ABOUT_TO_BE_SHOWN failed\n");

}

static osbool close_window(wimp_event_no event_code, wimp_block *block, toolbox_block *id_block, void *handle)
{
  destroy_window();
  return TRUE;

}

static void destroy_window(void)
{
  event_deregister_wimp_handler(event_ANY,
			wimp_REDRAW_WINDOW_REQUEST, redraw_window, NULL);
  event_deregister_wimp_handler(event_ANY,
			wimp_CLOSE_WINDOW_REQUEST, close_window, NULL);
  toolbox_delete_object(0, main_window);

  main_window = NULL;

  if (spritearea) {

    free(spritearea);
    spritearea = NULL;
    }

}

osbool redraw_window(wimp_event_no event_code, wimp_block *block, toolbox_block *id_block, void *handle)
{
 	UNUSED(event_code);
	UNUSED(id_block);
	UNUSED(handle);
  osbool more;
  wimp_draw redraw;
  int x0, y0;
  int xeig, yeig;
  BBox extent;
   xwindow_get_wimp_handle(0, main_window, &w);
   redraw.w = w;
   xwimp_redraw_window(&redraw, &more);
	os_factors factors;
	os_read_mode_variable(os_CURRENT_MODE, os_MODEVAR_XEIG_FACTOR, &xeig);
	os_read_mode_variable(os_CURRENT_MODE, os_MODEVAR_YEIG_FACTOR, &yeig);

  x0 = redraw.box.xmin - redraw.xscroll;
  y0 = redraw.box.ymax - redraw.yscroll;
  xwindow_get_extent(0, id_block->this_obj, &extent);
    y0 += extent.ymin;

        factors.xmul=2 - xeig;
        factors.ymul=2 - yeig;
        factors.xdiv=1;
        factors.ydiv=1;
        while (more)
        {
	         xosspriteop_put_sprite_scaled(osspriteop_PTR, spritearea, (osspriteop_id)header, x0, y0, 0, &factors, NULL);
	         xwimp_get_rectangle(&redraw, &more);
	}

 return TRUE;
}

static int bitmap_convert(char *filename, int file_type, char *output_file,
				size_t resize_width, size_t resize_height)
{

#define ThrowWandException(wand) \
{ \
  char \
    *description; \
 \
  ExceptionType \
    severity; \
 \
  description=MagickGetException(wand,&severity); \
  (void) fprintf(stderr,"%s %s %lu %s\n",GetMagickModule(),description); \
  description=(char *) MagickRelinquishMemory(description); \
}

 MagickBooleanType status;
 MagickWand *mw = NULL;
// MagickWand *trans_mw = NULL;
 size_t length;
 os_error *error;
 char format[12];
 char imageSize[12];
 char cropSize[16];

 MagickWandGenesis();

 /* Create a wand */
 mw = NewMagickWand();

 /* Read the input image */
 status=MagickReadImage(mw,__unixify(filename, 0, NULL, 0 , 0));
 if (status == MagickFalse)
 {
   ThrowWandException(mw);

   if(mw) mw = DestroyMagickWand(mw);
     MagickWandTerminus();

     xhourglass_off();

   return 1;
 }
 else
   {
       switch(file_type)
    {
    case 0x69c : // BMP
      strcpy(format, "BMP");
      break;
    case 0xff0 : // TIFF
      strcpy(format, "TIFF");
      break;
    case 0xb60 : // PNG
      strcpy(format, "PNG");
      break;
    case 0xc85 : // Jpeg
      strcpy(format, "JPEG");
      break;
    case 0x695 : // GIF
      strcpy(format, "GIF");
      break;
    case 0xa66 : // WEBP
      strcpy(format, "WEBP");
      break;
    case 0xa61 : // QOI
      strcpy(format, "QOI");
      break;
    case 0x69e : // PBMPlus
      strcpy(format, "PPM");
      break;
    case 0xf98 : // PhotoShp
      strcpy(format, "PSD");
      break;
    case 0x69d : // Targa
      strcpy(format, "TGA");
      break;
     default:
      strcpy(format, "SPR");
    }
    /* Resize the image */
    if (resize_width > 0)
    {
      sprintf(imageSize, "%dx%d", (int)resize_width, (int)resize_height);
      sprintf(cropSize, "%dx%d+0+0", (int)resize_width, (int)resize_height);
      fprintf(stderr, "Image Size: %s\n", imageSize);

       status = MagickAdaptiveResizeImage(mw, resize_width, resize_height);
         MagickSetImageFormat(mw, format);
   if (strcmp(format, "SPR") == 0)
   {
     unsigned char *image = MagickGetImageBlob(mw, &length);

     size = length;
     width = MagickGetImageWidth(mw);
     height = MagickGetImageHeight(mw);
     spritearea = (osspriteop_area *)malloc(length + sizeof(int));
     error = xosspriteop_clear_sprites(osspriteop_USER_AREA, spritearea);
     if (error)
     {
	fprintf(stderr, "osspriteop_clear_sprites: 0x%x: %s\n",
	error->errnum, error->errmess);
	if (spritearea)
	{
	  free(spritearea);
	  spritearea = NULL;
	}
     }
     spritearea->size = length + sizeof(int);
     spritearea->sprite_count = 1;
     spritearea->first = 16;
     spritearea->used = length + sizeof(int);
     memcpy((char *) spritearea + 4, image, length);

     free(image);

     header = (osspriteop_id) ((char *) spritearea + spritearea->first);
   }
 else
   MagickWriteImage(mw, output_file);
    }
  else
  {
 /* write it */
   MagickSetImageFormat(mw, format);
   if (strcmp(format, "SPR") == 0)
   {
     unsigned char *image = MagickGetImageBlob(mw, &length);

     size = length;
     width = MagickGetImageWidth(mw);
     height = MagickGetImageHeight(mw);
     spritearea = (osspriteop_area *)malloc(length + sizeof(int));
     error = xosspriteop_clear_sprites(osspriteop_USER_AREA, spritearea);
     if (error)
     {
	fprintf(stderr, "osspriteop_clear_sprites: 0x%x: %s\n",
	error->errnum, error->errmess);
	if (spritearea)
	{
	  free(spritearea);
	  spritearea = NULL;
	}
     }
     spritearea->size = length + sizeof(int);
     spritearea->sprite_count = 1;
     spritearea->first = 16;
     spritearea->used = length + sizeof(int);
     memcpy((char *) spritearea + 4, image, length);

     free(image);

     header = (osspriteop_id) ((char *) spritearea + spritearea->first);
   }
 else
   MagickWriteImage(mw, output_file);
}
 /* Tidy up */
 if(mw) mw = DestroyMagickWand(mw);

// if(trans_mw) trans_mw = DestroyMagickWand(trans_mw);

 MagickWandTerminus();
}

 return 0;

}

/**
 * Handle saveas events
 */
osbool toolbox_saveas(bits event_code, toolbox_action *event,
		toolbox_block *id_block, void *handle)
  {

    os_error *error;
    char menu_title[32];

	UNUSED(event_code);
	UNUSED(event);
	UNUSED(handle);

        error = xmenu_get_title(0, id_block->parent_obj, menu_title, 32, NULL);
    	if (error) {
		fprintf(stderr, "xmenu_get_title: 0x%x: %s\n",
			error->errnum, error->errmess);
	}
          fprintf(stderr, "Menu Title: %s\n", menu_title);
          fprintf(stderr, "This object: %d \n", id_block->parent_cmp);

	  if (strcmp(menu_title, "ExportMenu") == 0)
	  {

	  switch(id_block->parent_cmp)
	  {
	    case 0:
	    xsaveas_set_file_type(0, id_block->this_obj, 0xc85);
	    break;
	    case 1:
	    xsaveas_set_file_type(0, id_block->this_obj, 0xb60);
	    break;

	    case 2:
	    xsaveas_set_file_type(0, id_block->this_obj, 0x695);
	    break;

	    case 3:
	    xsaveas_set_file_type(0, id_block->this_obj, 0xff0);
	    break;

	    case 4:
	    xsaveas_set_file_type(0, id_block->this_obj, 0x69c);
	    break;

	    case 5:
	    xsaveas_set_file_type(0, id_block->this_obj, 0xf98);
	    break;

	    case 6:
	    xsaveas_set_file_type(0, id_block->this_obj, 0x69e);
	    break;

	    case 7:
	    xsaveas_set_file_type(0, id_block->this_obj, 0x69d);
	    break;

	    case 8:
	    xsaveas_set_file_type(0, id_block->this_obj, 0xa66);
	    break;

	    case 9:
	    xsaveas_set_file_type(0, id_block->this_obj, 0xa61);
	    break;
            }

	}
        else
        {

          error = xsaveas_set_file_size(0, id_block->this_obj, size - sizeof(int));
    	  if (error) {
		fprintf(stderr, "xsaveas_set_file_size: 0x%x: %s\n",
			error->errnum, error->errmess);
	  }

          error = xsaveas_set_data_address(0, id_block->this_obj, (byte *) spritearea + sizeof(int), size,
          				(byte *) spritearea + sizeof(int), size);
       	  if (error) {
		fprintf(stderr, "xsaveas_set_data_address: 0x%x: %s\n",
			error->errnum, error->errmess);
	  }

        }

	return TRUE;
}
osbool toolbox_savetofile(bits event_code, toolbox_action *event, toolbox_block *id_block, void *handle)
{
  bits file_type;
  os_error *error;
  char buffer[256];

          error = xsaveas_get_file_type(0, id_block->this_obj, &file_type);
    	  if (error) {
		fprintf(stderr, "xsaveas_get_file_type: 0x%x: %s\n",
			error->errnum, error->errmess);
	  }

   fprintf(stderr, "File Type: %d\n", file_type);

	  if (file_type == 0xff9)
	    return TRUE;

          error = xsaveas_get_file_name(0, id_block->this_obj, buffer, sizeof(buffer), 0);
    	  if (error) {
		fprintf(stderr, "xsaveas_get_file_type: 0x%x: %s\n",
			error->errnum, error->errmess);
	  }
   fprintf(stderr, "File Name: %s\n", buffer);

	  xhourglass_on();

  	bitmap_convert(imagefile, file_type, __unixify(buffer, 0, NULL, 0, 0), 0, 0);
  	xosfile_set_type(buffer, file_type);
  	xsaveas_file_save_completed((1), id_block->this_obj, buffer);

   xhourglass_off();

  return TRUE;

}

osbool toolbox_actionbutton(bits event_code, toolbox_action *event, toolbox_block *id_block, void *handle)
{
	UNUSED(event_code);
	UNUSED(event);
	UNUSED(id_block);
	UNUSED(handle);

 char imageWidth[5];
 char imageHeight[5];
 toolbox_c selected;
 os_error *error;
 int button;

 fprintf(stderr, "Action:\n");

  xwritablefield_get_value(0, id_block->this_obj, 0x0a, imageWidth, 5, 0);
  xwritablefield_get_value(0, id_block->this_obj, 0x09, imageHeight, 5, 0);
  fprintf(stderr, "Width: %d Height: %d\n", atoi(imageWidth), atoi(imageHeight));

  bitmap_load(imagefile, atoi(imageWidth), atoi(imageHeight));

  return TRUE;

}

osbool update_menu(bits event_code, toolbox_action *event, toolbox_block *id_block, void *handle)
{
 	UNUSED(event_code);
	UNUSED(event);
	UNUSED(handle);

  os_error *error;
  char menu_title[32];
  char file_size[12];
  char imageWidth[5];
  char imageHeight[5];
  char processTime[12];

  sprintf(file_size, "%d", size);
  sprintf(imageWidth, "%d", width);
  sprintf(imageHeight, "%d", height);
  sprintf(processTime, "%fs", endTime - startTime);

        error = xwindow_get_title(0, id_block->this_obj, menu_title, 32, NULL);
    	if (error) {
		fprintf(stderr, "xwindow_get_title: 0x%x: %s\n",
			error->errnum, error->errmess);
	}
          fprintf(stderr, "Menu Title: %s\n", menu_title);
          fprintf(stderr, "This object: %d \n", id_block->this_cmp);

	  if (strcmp(menu_title, "Resize") == 0)
	  {
	    xdisplayfield_set_value(0, id_block->this_obj, 0x1b, file_size);
	    xdisplayfield_set_value(0, id_block->this_obj, 0x3, imageWidth);
	    xdisplayfield_set_value(0, id_block->this_obj, 0x4, imageHeight);
	    xdisplayfield_set_value(0, id_block->this_obj, 0x1d, processTime);
	    fprintf(stderr, "Resize:\n");

          }
  return TRUE;

}
/**
 * Handle quit events
 */
osbool toolbox_event_quit(bits event_code, toolbox_action *event,
		toolbox_block *id_block, void *handle)
{
	UNUSED(event_code);
	UNUSED(event);
	UNUSED(id_block);
	UNUSED(handle);

	toolbox_quit = 1;

	return TRUE;
}

/**
 * Handle toolbox errors
 */
osbool toolbox_event_error(bits event_code, toolbox_action *event,
		toolbox_block *id_block, void *handle)
{
	toolbox_action_error_block *error =
			(toolbox_action_error_block *) event;

	UNUSED(event_code);
	UNUSED(id_block);
	UNUSED(handle);

	fprintf(stderr, "toolbox error: 0x%x: %s\n",
		error->errnum, error->errmess);

	return TRUE;
}

/**
 * Register message handlers
 */
void register_message_handlers(void)
{
	osbool success;

	success = event_register_message_handler(message_QUIT,
			wimp_message_quit, NULL);
	if (success == FALSE)
		fprintf(stderr, "registering message_QUIT handler failed\n");

	success = event_register_message_handler(message_DATA_LOAD,
			wimp_message_dataload, NULL);
	if (success == FALSE)
		fprintf(stderr,
			"registering message_DATA_LOAD handler failed\n");
	success = event_register_message_handler(message_DATA_OPEN,
			wimp_message_dataload, NULL);
	if (success == FALSE)
		fprintf(stderr,
			"registering message_DATA_OPEN handler failed\n");
}

/**
 * Handle quit messages
 */
osbool wimp_message_quit(wimp_message *message, void *handle)
{
	UNUSED(message);
	UNUSED(handle);

	toolbox_quit = 1;

	return TRUE;
}

/**
 * Handle dataload messages
 */
osbool wimp_message_dataload(wimp_message *message, void *handle)
{
	os_error *error;
	wimp_message_data_xfer *dl = &message->data.data_xfer;

	UNUSED(handle);

  if (dl->file_type > 0xfff)
    return FALSE;

    if (dl->w != wimp_ICON_BAR)
    {
    switch(dl->file_type)
    {
    case 0x69c : // BMP
    case 0xff0 : // TIFF
    case 0xaad : // SVG
    case 0xb60 : // PNG
    case 0xb61 : // XBitMap
    case 0x301 : // XPixMap
    case 0xf83 : // MNG
    case 0xf78 : // JNG
    case 0xc85 : // Jpeg
    case 0x69d : // Targa
    case 0xfd5 : // Pict2
    case 0x697 : // PCX
    case 0x69e : // PBMPlus
    case 0xfc9 : // SunRastr
    case 0x699 : // MTV
    case 0x692 : // IMG
    case 0x691 : // Degas
    case 0x6a2 : // ColoRIX
    case 0x698 : // QRT
    case 0xf98 : // PhotoShp
    case 0xb2d : // AdobeIll
    case 0x6a6 : // WPG
    case 0xb2f : // WinMeta
    case 0x695 : // GIF
    case 0xa66 : // WEBP
    case 0xa61 : // QOI
    case 0xb2b : // CGM
    case 0xce4 : // DVI
    case 0x132 : // ICO
     break;
    default:
      return FALSE;
    }
    }
	message->action = message_DATA_LOAD_ACK;
	message->your_ref = message->my_ref;

	error = xwimp_send_message(wimp_USER_MESSAGE, message, message->sender);
	if (error != NULL) {
		fprintf(stderr, "wimp_send_message: 0x%x: %s\n",
			error->errnum, error->errmess);
	  }
	strcpy(imagefile, dl->file_name);
	bitmap_load(imagefile, 0, 0);

	return TRUE;
}
